/*
 * Copyright 1993-2012 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */


/*
    Parallel reduction

    This sample shows how to perform a reduction operation on an array of values
    to produce a single value.

    Reductions are a very common computation in parallel algorithms.  Any time
    an array of values needs to be reduced to a single value using a binary
    associative operator, a reduction can be used.  Example applications include
    statistics computaions such as mean and standard deviation, and image
    processing applications such as finding the total luminance of an
    image.

    This code performs sum reductions, but any associative operator such as
    min() or max() could also be used.

    It assumes the input size is a power of 2.

    COMMAND LINE ARGUMENTS

    "--shmoo":         Test performance for 1 to 32M elements with each of the 7 different kernels
    "--n=<N>":         Specify the number of elements to reduce (default 1048576)
    "--threads=<N>":   Specify the number of threads per block (default 128)
    "--kernel=<N>":    Specify which kernel to run (0-6, default 6)
    "--maxblocks=<N>": Specify the maximum number of thread blocks to launch (kernel 6 only, default 64)
    "--cpufinal":      Read back the per-block results and do final sum of block sums on CPU (default false)
    "--cputhresh=<N>": The threshold of number of blocks sums below which to perform a CPU final reduction (default 1)
    "-type=<T>":       The datatype for the reduction, where T is "int", "float", or "double" (default int)
*/

// CUDA Runtime
#include <cuda_runtime.h>

// Utilities and system includes
#include <helper_cuda.h>
#include <helper_functions.h>
#include <algorithm>

// includes, project
#include "reduction.h"
#include "half.h"
#include "./inc/gqd_type.h"

using half_float::half;

// vector type for half
typedef unsigned short ushort;

// Input type
enum ReduceType	
{
    REDUCE_HALF,
    REDUCE_HALF2,
    REDUCE_HALF4,
    REDUCE_FLOAT,
	REDUCE_FLOAT2,
	REDUCE_FLOAT4,
    REDUCE_DOUBLE,
    REDUCE_DOUBLE2
};

// Use an accumulator with extra pecision
enum ExtraPrecAcc
{
	ACC_NM,			// Standard accumulator
    ACC_DD,			// Use doudle-double accumulator
	ACC_QD			// Use quad-double accumulator
};

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
template <class Tin>
bool runTest(int argc, char **argv, ReduceType datatype, ExtraPrecAcc acctype);

#define MAX_BLOCK_DIM_SIZE 65535

#ifdef WIN32
#define strcasecmp strcmpi
#endif

extern "C"
bool isPow2(unsigned int x)
{
    return ((x&(x-1))==0);
}
#define SPLIT "##########################################################################################################\n"

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int
main(int argc, char **argv)
{
    printf("%s Starting...\n\n", argv[0]);

    char *typeChoice;
    getCmdLineArgumentString(argc, (const char **) argv, "typeIn", &typeChoice);

    if (0 == typeChoice){
        typeChoice = (char *)malloc(8 * sizeof(char));
        strcpy(typeChoice, "float");
    }

    ReduceType datatypeIn = REDUCE_FLOAT;

    if (!strcasecmp(typeChoice, "half"))
        datatypeIn = REDUCE_HALF;
    else if (!strcasecmp(typeChoice, "half2"))
        datatypeIn = REDUCE_HALF2;
    else if (!strcasecmp(typeChoice, "half4"))
        datatypeIn = REDUCE_HALF4;
	else if (!strcasecmp(typeChoice, "float"))
        datatypeIn = REDUCE_FLOAT;
	else if (!strcasecmp(typeChoice, "float2"))
        datatypeIn = REDUCE_FLOAT2;
	else if (!strcasecmp(typeChoice, "float4"))
        datatypeIn = REDUCE_FLOAT4;
    else if (!strcasecmp(typeChoice, "double"))
        datatypeIn = REDUCE_DOUBLE;
    else if (!strcasecmp(typeChoice, "double2"))
        datatypeIn = REDUCE_DOUBLE2;
    else{
		fprintf(stderr,"ERROR: Unknown Type %s \n",typeChoice);
		exit(EXIT_FAILURE);
	}

    char *ExtraPrecAccChoice;
    getCmdLineArgumentString(argc, (const char **) argv, "ac", &ExtraPrecAccChoice);

    if (0 == ExtraPrecAccChoice){
        ExtraPrecAccChoice = (char *)malloc(8 * sizeof(char));
        strcpy(ExtraPrecAccChoice, "Normal");
    }
    ExtraPrecAcc AccType = ACC_NM;

    if (!strcasecmp(ExtraPrecAccChoice, "Normal"))
        AccType = ACC_NM;
    else if (!strcasecmp(ExtraPrecAccChoice, "DD"))
        AccType = ACC_DD;
    else if (!strcasecmp(ExtraPrecAccChoice, "QD"))
        AccType = ACC_QD;
	

	if ((datatypeIn != REDUCE_DOUBLE) && (AccType != ACC_NM)){
		fprintf(stderr,"WARNING: Reduction an array of type %s with an accumulator of type %s is not yet \
					   supported denoting the accumulator to a Normal accumulator\n", typeChoice, ExtraPrecAccChoice);
		strcpy(ExtraPrecAccChoice, "Normal");
		AccType = ACC_NM;
	}

    cudaDeviceProp deviceProp;
    deviceProp.major = 1;
    deviceProp.minor = 0;
    int minimumComputeVersion = 10;

    if (datatypeIn == REDUCE_DOUBLE){
        deviceProp.minor = 3;
        minimumComputeVersion = 13;
    }

    int dev;

    dev = findCudaDevice(argc, (const char **)argv);

    checkCudaErrors(cudaGetDeviceProperties(&deviceProp, dev));

    if ((deviceProp.major * 10 + deviceProp.minor) >= minimumComputeVersion){
        printf("Using Device %d: %s\n\n", dev, deviceProp.name);
        checkCudaErrors(cudaSetDevice(dev));
    }else {
        printf("Error: the selected device does not support the minimum compute capability of %d.%d.\n\n",
               minimumComputeVersion / 10, minimumComputeVersion % 10);

        cudaDeviceReset();
        exit(EXIT_FAILURE);
    }

    printf("Reducing array of type %s\n", typeChoice);
	printf("With an accumulator of type %s\n\n", ExtraPrecAccChoice);

    bool bResult = false;
    switch (datatypeIn)
    {
        default:
        case REDUCE_HALF:
            bResult = runTest<ushort>(argc, argv, datatypeIn, AccType);
            break;
        case REDUCE_HALF2:
            bResult = runTest<ushort2>(argc, argv, datatypeIn, AccType);
            break;
        case REDUCE_HALF4:
            bResult = runTest<ushort4>(argc, argv, datatypeIn, AccType);
            break;

        case REDUCE_FLOAT:
            bResult = runTest<float>(argc, argv, datatypeIn, AccType);
            break;
        case REDUCE_FLOAT2:
            bResult = runTest<float2>(argc, argv, datatypeIn, AccType);
            break;
        case REDUCE_FLOAT4:
            bResult = runTest<float4>(argc, argv, datatypeIn, AccType);
            break;

        case REDUCE_DOUBLE:
            bResult = runTest<double>(argc, argv, datatypeIn, AccType);
            break;
		case REDUCE_DOUBLE2:
            bResult = runTest<double2>(argc, argv, datatypeIn, AccType);
            break;
	}
	
    cudaDeviceReset();
    printf(bResult ? "Test passed\n" : "Test failed!\n");
}

////////////////////////////////////////////////////////////////////////////////
//! Compute sum reduction on CPU
//! We use Kahan summation for an accurate sum of large arrays.
//! http://en.wikipedia.org/wiki/Kahan_summation_algorithm
//!
//! @param data       pointer to input data
//! @param size       number of input data elements
////////////////////////////////////////////////////////////////////////////////
template<typename Tin, typename Tout>
Tout reduceCPU(Tin *data, int size, ReduceType datatype)
{
    Tout sum; 
    Tout c = (Tout)0.0;
	half *h;
	float *f;
	double *d;
	h = (half *)data;
	f = (float *)data;
	d = (double *)data;

	switch(datatype){
		case REDUCE_HALF4:
			size*=2;
		case REDUCE_HALF2:
			size*=2;
		case REDUCE_HALF:
			sum = (Tout) h[0]; 
		   for (int i = 1; i < size; i++) {
		      Tout y = (Tout)h[i] - c;
		       Tout t = sum + y;
			     c = (t - sum) - y;
			   sum = t;
		   }
		   break;

		case REDUCE_FLOAT4:
			size*=2;
		case REDUCE_FLOAT2:
			size*=2;
		case REDUCE_FLOAT:
			sum = (Tout) f[0]; 
		   for (int i = 1; i < size; i++) {
		      Tout y = (Tout)f[i] - c;
		       Tout t = sum + y;
			     c = (t - sum) - y;
			   sum = t;
		   }
		   break;

		case REDUCE_DOUBLE2:
			size*=2;
		case REDUCE_DOUBLE:
			sum = (Tout) d[0]; 
		   for (int i = 1; i < size; i++) {
		      Tout y = (Tout)d[i] - c;
		       Tout t = sum + y;
			     c = (t - sum) - y;
			   sum = t;
		   }
		   break;
	}

    return sum;
}

unsigned int nextPow2(unsigned int x)
{
    --x;
    x |= x >> 1;
    x |= x >> 2;
    x |= x >> 4;
    x |= x >> 8;
    x |= x >> 16;
    return ++x;
}

#ifndef MIN
#define MIN(x,y) ((x < y) ? x : y)
#endif

////////////////////////////////////////////////////////////////////////////////
// Compute the number of threads and blocks to use for the given reduction kernel
// We set threads / block to the minimum of maxThreads and n/2. We observe the 
// maximum specified number of blocks, because each thread in that kernel can 
// process a variable number of elements.
////////////////////////////////////////////////////////////////////////////////
void getNumBlocksAndThreads(int n, int maxBlocks, int maxThreads, int &blocks, int &threads)
{

    //get device capability, to avoid block/grid size excceed the upbound
    cudaDeviceProp prop;
    int device;
    checkCudaErrors(cudaGetDevice(&device));
    checkCudaErrors(cudaGetDeviceProperties(&prop, device));

    threads = (n < maxThreads*2) ? nextPow2((n + 1)/ 2) : maxThreads;
    blocks = (n + (threads * 2 - 1)) / (threads * 2);
 
	printf("Threads: %d \n", threads);
	printf("Blocks: %d \n", blocks);
	printf("maxGridSize: %d \n", prop.maxGridSize[0]);
	printf("maxThreadsPerBlock: %d \n", prop.maxThreadsPerBlock);

    if ((long long int)threads*blocks > (long long int)prop.maxGridSize[0] * prop.maxThreadsPerBlock){
        printf("n is too large, please choose a smaller number!\n");
    }

    if (blocks > prop.maxGridSize[0]){
        printf("Grid size <%d> excceeds the device capability <%d>, set block size as %d (original %d)\n",
               blocks, prop.maxGridSize[0], threads*2, threads);

        blocks /= 2;
        threads *= 2;
    }

    blocks = MIN(maxBlocks, blocks);
}

////////////////////////////////////////////////////////////////////////////////
// This function performs a reduction of the input data multiple times and
// measures the average reduction time.
////////////////////////////////////////////////////////////////////////////////
template <typename Tin, typename Tout>
Tout benchmarkReduce(int  n,
                  int  numThreads,
                  int  numBlocks,
                  int  maxThreads,
                  int  maxBlocks,
                  int  testIterations,
                  StopWatchInterface *timer,
                  Tout *h_odata,
                  Tin *d_idata,
                  Tout *d_odata,
				  ExtraPrecAcc acctype){
    Tout gpu_result;

    for (int i = 0; i < testIterations; ++i)
    {
        gpu_result=0;

        cudaDeviceSynchronize();
        sdkStartTimer(&timer);

        // execute the kernel
		if (acctype == ACC_NM)
			reduce<Tin, Tout>(n, numThreads, numBlocks, d_idata, d_odata);
		else if (acctype == ACC_DD)
			reduce_extra_prec<gdd_real>(n, numThreads, numBlocks, (double*)d_idata, (double*)d_odata);
		else if (acctype == ACC_QD)
			reduce_extra_prec<gqd_real>(n, numThreads, numBlocks, (double*)d_idata, (double*)d_odata);

 
        // check if kernel execution generated an error
        getLastCudaError("Kernel execution failed");

        // sum partial sums from each block on CPU
        // copy result from device to host
        checkCudaErrors(cudaMemcpy(h_odata, d_odata, numBlocks*sizeof(Tout), cudaMemcpyDeviceToHost));

        for (int i=0; i<numBlocks; i++)
           gpu_result = gpu_result + h_odata[i];

 		cudaDeviceSynchronize();
        sdkStopTimer(&timer);
    }

    return gpu_result;
}

ushort getrand(ushort t){
	half f;
	f = half((rand() & 0xFF) / (float)(RAND_MAX));
	return f.data_;
}
ushort2 getrand(ushort2 t){
	t.x = getrand((ushort)0.);
	t.y = getrand((ushort)0.);
	return t;
}
ushort4 getrand(ushort4 t){
	t.x = getrand((ushort)0.);
	t.y = getrand((ushort)0.);
	t.z = getrand((ushort)0.);
	t.w = getrand((ushort)0.);
	return t;
}

float getrand(float t){
	return (float)((rand() & 0xFF) / (float)(RAND_MAX));
}
float2 getrand(float2 t){
	float2 f;
	f.x = getrand(float(0.));
	f.y = getrand(float(0.));
	return f;
}
float4 getrand(float4 t){
	float4 f;
	f.x = getrand(float(0.));
	f.y = getrand(float(0.));
	f.z = getrand(float(0.));
	f.w = getrand(float(0.));
	return f;
}

double getrand(double t){
	return (float)((rand() & 0xFF) / (float)(RAND_MAX));
}
double2 getrand(double2 t){
	double2 f;
	f.x = getrand(double(0.));
	f.y = getrand(double(0.));
	return f;
}

////////////////////////////////////////////////////////////////////////////////
// The main function whihc runs the reduction test.
////////////////////////////////////////////////////////////////////////////////
template <typename Tin, typename Tout>
bool testWithDataType(int size, int maxThreads, int maxBlocks, ReduceType datatype, ExtraPrecAcc acctype)
{
    // create random input data on CPU
    unsigned int bytes = size * sizeof(Tin);
	Tin *h_idata = (Tin *) malloc(bytes);

	for (int i=0; i<size; i++)
			h_idata[i] = getrand(h_idata[0]);

    int numBlocks = 0;
    int numThreads = 0;
    getNumBlocksAndThreads(size, maxBlocks, maxThreads, numBlocks, numThreads);


    // allocate mem for the result on host side
    Tout *h_odata = (Tout *) malloc(numBlocks*sizeof(Tout));

    printf("%d blocks\n\n", numBlocks);

    // allocate device memory and data
    Tin *d_idata = NULL;
    Tout *d_odata = NULL;

    checkCudaErrors(cudaMalloc((void **) &d_idata, bytes));
    checkCudaErrors(cudaMalloc((void **) &d_odata, numBlocks*sizeof(Tout)));

    // copy data directly to device memory
    checkCudaErrors(cudaMemcpy(d_idata, h_idata, bytes, cudaMemcpyHostToDevice));
    checkCudaErrors(cudaMemcpy(d_odata, h_idata, numBlocks*sizeof(Tout), cudaMemcpyHostToDevice));
	
    // warm-up
	if (acctype == ACC_NM)
		reduce<Tin, Tout>(size, numThreads, numBlocks, d_idata, d_odata);
	else if (acctype == ACC_DD)
		reduce_extra_prec<gdd_real>(size, numThreads, numBlocks, (double*)d_idata, (double*)d_odata);
	else if (acctype == ACC_QD)
		reduce_extra_prec<gqd_real>(size, numThreads, numBlocks, (double*)d_idata, (double*)d_odata);

    int testIterations = 100;

    StopWatchInterface *timer = 0;
    sdkCreateTimer(&timer);

    Tout gpu_result;

 	gpu_result = benchmarkReduce<Tin, Tout>(size, numThreads, numBlocks, maxThreads, maxBlocks,
                                        testIterations, timer, h_odata, d_idata, d_odata, acctype);


    double reduceTime = sdkGetAverageTimerValue(&timer) * 1e-3;
    printf("Reduction, Throughput = %.4f GB/s, Time = %.5f s, Size = %u Elements, NumDevsUsed = %d, Workgroup = %u\n",
               1.0e-9 * ((double)bytes)/reduceTime, reduceTime, size, 1, numThreads);

    // compute reference solution
	Tout cpu_result;
	cpu_result = reduceCPU<Tin, Tout>(h_idata, size, datatype);
		
    double threshold = 1e-12;
    double diff = 0;

    printf("\nGPU result = %f\n", gpu_result);
    printf("CPU result = %f\n\n", cpu_result);
    threshold = 1e-8 * size;

    diff = fabs((double)gpu_result - (double)cpu_result);

    // cleanup
    sdkDeleteTimer(&timer);
    free(h_idata);
    free(h_odata);

    checkCudaErrors(cudaFree(d_idata));
    checkCudaErrors(cudaFree(d_odata));

    return (diff < threshold);
}


////////////////////////////////////////////////////////////////////////////////
// The main function whihc runs the reduction test.
////////////////////////////////////////////////////////////////////////////////
template <class Tin>
bool
runTest(int argc, char **argv, ReduceType datatype, ExtraPrecAcc acctype)
{
    int size = 1<<25;    // number of elements to reduce
    int maxThreads = 256;  // number of threads per block
    int maxBlocks = 64;

    if (checkCmdLineFlag(argc, (const char **) argv, "n"))
    {
        size = getCmdLineArgumentInt(argc, (const char **) argv, "n");
    }

    if (checkCmdLineFlag(argc, (const char **) argv, "threads"))
    {
        maxThreads = getCmdLineArgumentInt(argc, (const char **) argv, "threads");
    }

    if (checkCmdLineFlag(argc, (const char **) argv, "maxblocks"))
    {
        maxBlocks = getCmdLineArgumentInt(argc, (const char **) argv, "maxblocks");
    }

	// Adjust size according to the type
	if ((datatype == REDUCE_HALF2)||(datatype == REDUCE_FLOAT2)||(datatype == REDUCE_DOUBLE2)){
		size = size/2;
	}
	if ((datatype == REDUCE_HALF4)||(datatype == REDUCE_FLOAT4)){
		size = size/4;
	}

    printf("%d elements\n", size);
    printf("%d threads (max)\n", maxThreads);
	printf(SPLIT);
	printf("FLOAT Accumulator \n");
	testWithDataType<Tin, float>(size, maxThreads, maxBlocks, datatype, acctype);

	printf(SPLIT);
	printf("DOUBLE Accumulator \n");
	return testWithDataType<Tin, double>(size, maxThreads, maxBlocks, datatype, acctype);


 }
