/*
 * Copyright 1993-2012 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */

/*
    Parallel reduction kernels
*/

#ifndef _REDUCE_KERNEL_H_
#define _REDUCE_KERNEL_H_

#include <stdio.h>
#include "./inc/gqd_type.h"
#include "./inc/gqd.cu"


// Utility class used to avoid linker errors with extern
// unsized shared memory arrays with templated type
template<class T>
struct SharedMemory
{
    __device__ inline operator       T *()
    {
        extern __shared__ int __smem[];
        return (T *)__smem;
    }

    __device__ inline operator const T *() const
    {
        extern __shared__ int __smem[];
        return (T *)__smem;
    }
};


// specialize for double to avoid unaligned memory
// access compile errors
template<>
struct SharedMemory<double>
{
    __device__ inline operator       double *()
    {
        extern __shared__ double __smem_d[];
        return (double *)__smem_d;
    }

    __device__ inline operator const double *() const
    {
        extern __shared__ double __smem_d[];
        return (double *)__smem_d;
    }
};

typedef unsigned short ushort;
// With convertion
__device__ inline float  convert(ushort x){return __half2float(x);}
__device__ inline float  convert(ushort2 x){return __half2float(x.x)+__half2float(x.y);}
__device__ inline float  convert(ushort4 x){return __half2float(x.x)+__half2float(x.y)+__half2float(x.z)+__half2float(x.w);}

__device__ inline float  convert(float x){return x;}
__device__ inline float  convert(float2 x){return x.x+x.y;}
__device__ inline float  convert(float4 x){return x.x+x.y+x.z+x.w;}

__device__ inline double convert(double x){return x;}
__device__ inline double convert(double2 x){return x.x+x.y;}

__device__ inline void myInit(float *x){*x=0.;}
__device__ inline void myInit(double *x){*x=0.;}
__device__ inline void myInit(double2 *x){(*x).x=0.;(*x).y=0.;}

/*
    Parallel sum reduction using shared memory
    - takes log(n) steps for n input elements
    - uses n threads
    - only works for power-of-2 arrays
*/

/*
    This version adds multiple elements per thread sequentially.  This reduces the overall
    cost of the algorithm while keeping the work complexity O(n) and the step complexity O(log n).
    (Brent's Theorem optimization)

    Note, this kernel needs a minimum of 64*sizeof(T) bytes of shared memory.
    In other words if blockSize <= 32, allocate 64*sizeof(T) bytes.
    If blockSize > 32, allocate blockSize*sizeof(T) bytes.
*/
template <typename Tin, typename Tout, unsigned int blockSize, bool nIsPow2>
__global__ void
reduce0(Tin *g_idata, Tout *g_odata, unsigned int n)
{
    Tout *sdata = SharedMemory<Tout>();

    // perform first level of reduction,
    // reading from global memory, writing to shared memory
    unsigned int tid = threadIdx.x;
    unsigned int i = blockIdx.x*blockSize*2 + threadIdx.x;
    unsigned int gridSize = blockSize*2*gridDim.x;

    Tout mySum;
	myInit(&mySum);

    // we reduce multiple elements per thread.  The number is determined by the
    // number of active thread blocks (via gridDim).  More blocks will result
    // in a larger gridSize and therefore fewer elements per thread
    while (i < n)
    {
		mySum = mySum + convert(g_idata[i]);

        // ensure we don't read out of bounds -- this is optimized away for powerOf2 sized arrays
        if (nIsPow2 || i + blockSize < n)
			mySum = mySum + convert(g_idata[i+blockSize]);
 		

        i += gridSize;
    }

    // each thread puts its local sum into shared memory
    sdata[tid] = mySum;
    __syncthreads();


    // do reduction in shared mem
    if (blockSize >= 512)
    {
        if (tid < 256)
            sdata[tid] = mySum = mySum + sdata[tid + 256];

        __syncthreads();
    }

    if (blockSize >= 256)
    {
        if (tid < 128)
            sdata[tid] = mySum = mySum + sdata[tid + 128];

        __syncthreads();
    }

    if (blockSize >= 128)
    {
        if (tid <  64)
            sdata[tid] = mySum = mySum + sdata[tid +  64];

        __syncthreads();
    }

    if (tid < 32)
    {
        // now that we are using warp-synchronous programming (below)
        // we need to declare our shared memory volatile so that the compiler
        // doesn't reorder stores to it and induce incorrect behavior.
		 volatile Tout *smem = sdata;

        if (blockSize >=  64)
            smem[tid] = mySum = mySum + smem[tid + 32];

        if (blockSize >=  32)
            smem[tid] = mySum = mySum + smem[tid + 16];

        if (blockSize >=  16)
            smem[tid] = mySum = mySum + smem[tid +  8];

        if (blockSize >=   8)
            smem[tid] = mySum = mySum + smem[tid +  4];

        if (blockSize >=   4)
            smem[tid] = mySum = mySum + smem[tid +  2];

        if (blockSize >=   2)
            smem[tid] = mySum = mySum + smem[tid +  1];
    }

    // write result for this block to global mem
    if (tid == 0)
        g_odata[blockIdx.x] = sdata[0];
		
}
 

/*
    This version adds multiple elements per thread sequentially.  This reduces the overall
    cost of the algorithm while keeping the work complexity O(n) and the step complexity O(log n).
    (Brent's Theorem optimization)

    Note, this kernel needs a minimum of 64*sizeof(T) bytes of shared memory.
    In other words if blockSize <= 32, allocate 64*sizeof(T) bytes.
    If blockSize > 32, allocate blockSize*sizeof(T) bytes.
*/
template <typename Acc, unsigned int blockSize, bool nIsPow2>
__global__ void
reduce0_extra_prec(double *g_idata, double *g_odata, unsigned int n){
    Acc *sdata = SharedMemory<Acc>();

    // perform first level of reduction,
    // reading from global memory, writing to shared memory
    unsigned int tid = threadIdx.x;
    unsigned int i = blockIdx.x*blockSize*2 + threadIdx.x;
    unsigned int gridSize = blockSize*2*gridDim.x;

    Acc mySum;
	mySum.x=0;
	mySum.y=0;


    // we reduce multiple elements per thread.  The number is determined by the
    // number of active thread blocks (via gridDim).  More blocks will result
    // in a larger gridSize and therefore fewer elements per thread
    while (i < n){
		mySum = mySum + g_idata[i];
        // ensure we don't read out of bounds -- this is optimized away for powerOf2 sized arrays
        if (nIsPow2 || i + blockSize < n)
			mySum = mySum + g_idata[i+blockSize];
 		
        i += gridSize;
    }

    // each thread puts its local sum into shared memory
    sdata[tid] = mySum;
    __syncthreads();


    // do reduction in shared mem
    if (blockSize >= 512){
        if (tid < 256)
            sdata[tid] = mySum = mySum + sdata[tid + 256];
        __syncthreads();
    }

    if (blockSize >= 256){
        if (tid < 128)
            sdata[tid] = mySum = mySum + sdata[tid + 128];
        __syncthreads();
    }

    if (blockSize >= 128){
        if (tid <  64)
            sdata[tid] = mySum = mySum + sdata[tid +  64];
        __syncthreads();
    }

    if (tid < 32){
        // now that we are using warp-synchronous programming (below)
        // we need to declare our shared memory volatile so that the compiler
        // doesn't reorder stores to it and induce incorrect behavior.
		 Acc *smem = sdata;

        if (blockSize >=  64){
            smem[tid] = mySum = mySum + smem[tid + 32];__syncthreads();
		}
        if (blockSize >=  32){
            smem[tid] = mySum = mySum + smem[tid + 16];__syncthreads();
		}
        if (blockSize >=  16){
            smem[tid] = mySum = mySum + smem[tid +  8];__syncthreads();
		}
        if (blockSize >=   8){
            smem[tid] = mySum = mySum + smem[tid +  4];__syncthreads();
		}
        if (blockSize >=   4){
            smem[tid] = mySum = mySum + smem[tid +  2];__syncthreads();
		}
        if (blockSize >=   2){
            smem[tid] = mySum = mySum + smem[tid +  1];__syncthreads();
		}
    }

    // write result for this block to global mem
    if (tid == 0){
        g_odata[blockIdx.x] = mySum.x;
	}
}
 

extern "C"
bool isPow2(unsigned int x);


////////////////////////////////////////////////////////////////////////////////
// Wrapper function for kernel launch
////////////////////////////////////////////////////////////////////////////////
template <typename Tin, typename Tout>
void
reduce(int size, int threads, int blocks, Tin *d_idata, Tout *d_odata)
{
    dim3 dimBlock(threads, 1, 1);
    dim3 dimGrid(blocks, 1, 1);

    // when there is only one warp per block, we need to allocate two warps
    // worth of shared memory so that we don't index shared memory out of bounds
    int smemSize = (threads <= 32) ? 2 * threads * sizeof(Tout) : threads * sizeof(Tout);

            if (isPow2(size))
            {
                switch (threads)
                {
                    case 512:
                        reduce0<Tin, Tout, 512, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 256:
                        reduce0<Tin, Tout, 256, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 128:
                        reduce0<Tin, Tout, 128, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 64:
                        reduce0<Tin, Tout,  64, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 32:
                        reduce0<Tin, Tout,  32, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 16:
                        reduce0<Tin, Tout,  16, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  8:
                        reduce0<Tin, Tout,   8, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  4:
                        reduce0<Tin, Tout,   4, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  2:
                        reduce0<Tin, Tout,   2, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  1:
                        reduce0<Tin, Tout,   1, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                }
            }
            else
            {
                switch (threads)
                {
                    case 512:
                        reduce0<Tin, Tout, 512, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 256:
                        reduce0<Tin, Tout, 256, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 128:
                        reduce0<Tin, Tout, 128, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 64:
                        reduce0<Tin, Tout,  64, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 32:
                        reduce0<Tin, Tout,  32, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 16:
                        reduce0<Tin, Tout,  16, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  8:
                        reduce0<Tin, Tout,   8, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  4:
                        reduce0<Tin, Tout,   4, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  2:
                        reduce0<Tin, Tout,   2, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  1:
                        reduce0<Tin, Tout,   1, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                }
            }
 }

template <typename Acc>
void
reduce_extra_prec(int size, int threads, int blocks, double *d_idata, double *d_odata){
    dim3 dimBlock(threads, 1, 1);
    dim3 dimGrid(blocks, 1, 1);

    // when there is only one warp per block, we need to allocate two warps
    // worth of shared memory so that we don't index shared memory out of bounds
    int smemSize = (threads <= 32) ? 2 * threads * sizeof(Acc) : threads * sizeof(Acc);

            if (isPow2(size))
            {
                switch (threads)
                {
                    case 512:
                        reduce0_extra_prec<Acc, 512, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 256:
                        reduce0_extra_prec<Acc, 256, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 128:
                        reduce0_extra_prec<Acc, 128, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 64:
                        reduce0_extra_prec<Acc,  64, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 32:
                        reduce0_extra_prec<Acc,  32, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 16:
                        reduce0_extra_prec<Acc,  16, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  8:
                        reduce0_extra_prec<Acc,   8, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  4:
                        reduce0_extra_prec<Acc,   4, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  2:
                        reduce0_extra_prec<Acc,   2, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  1:
                        reduce0_extra_prec<Acc,   1, true><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                }
            }
            else
            {
                switch (threads)
                {
                    case 512:
                        reduce0_extra_prec<Acc, 512, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 256:
                        reduce0_extra_prec<Acc, 256, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 128:
                        reduce0_extra_prec<Acc, 128, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 64:
                        reduce0_extra_prec<Acc,  64, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 32:
                        reduce0_extra_prec<Acc,  32, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case 16:
                        reduce0_extra_prec<Acc,  16, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  8:
                        reduce0_extra_prec<Acc,   8, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  4:
                        reduce0_extra_prec<Acc,   4, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  2:
                        reduce0_extra_prec<Acc,   2, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                    case  1:
                        reduce0_extra_prec<Acc,   1, false><<< dimGrid, dimBlock, smemSize >>>(d_idata, d_odata, size); break;
                }
            }
 }

// Instantiate the reduction function for various types
template void reduce<ushort, float>(int size, int threads, int blocks, ushort *d_idata, float *d_odata);
template void reduce<ushort, double>(int size, int threads, int blocks, ushort *d_idata, double *d_odata);
template void reduce<ushort2, float>(int size, int threads, int blocks, ushort2 *d_idata, float *d_odata);
template void reduce<ushort2, double>(int size, int threads, int blocks, ushort2 *d_idata, double *d_odata);
template void reduce<ushort4, float>(int size, int threads, int blocks, ushort4 *d_idata, float *d_odata);
template void reduce<ushort4, double>(int size, int threads, int blocks, ushort4 *d_idata, double *d_odata);
template void reduce<float, float>(int size, int threads, int blocks, float *d_idata, float *d_odata);
template void reduce<float, double>(int size, int threads, int blocks, float *d_idata, double *d_odata);
template void reduce<float2, float>(int size, int threads, int blocks, float2 *d_idata, float *d_odata);
template void reduce<float2, double>(int size, int threads, int blocks, float2 *d_idata, double *d_odata);
template void reduce<float4, float>(int size, int threads, int blocks, float4 *d_idata, float *d_odata);
template void reduce<float4, double>(int size, int threads, int blocks, float4 *d_idata, double *d_odata);
template void reduce<double, float>(int size, int threads, int blocks, double *d_idata, float *d_odata);
template void reduce<double, double>(int size, int threads, int blocks, double *d_idata, double *d_odata);
template void reduce<double2, float>(int size, int threads, int blocks, double2 *d_idata, float *d_odata);
template void reduce<double2, double>(int size, int threads, int blocks, double2 *d_idata, double *d_odata);

template void reduce_extra_prec<gqd_real>(int size, int threads, int blocks, double *d_idata, double *d_odata);
template void reduce_extra_prec<gdd_real>(int size, int threads, int blocks, double *d_idata, double *d_odata);

#endif // #ifndef _REDUCE_KERNEL_H_
